#include "../../include/error.h"
#include "../../include/pool.h"
#include "../../include/util.h"
#include "../../include/ctx.h"

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

bool lm_ctx_remove(lm_ctx_t *ctx, lm_pkg_t *pkg, lm_ctx_remove_callback_t callback, void *data){
  if(NULL == ctx && NULL == pkg){
    lm_error_set(LM_ERR_ArgNULL);
    return false;
  }

  if(NULL == ctx->root){
    lm_error_set(LM_ERR_CtxRootNULL);
    return false;
  }

  if(!lm_ctx_database_is_installed(ctx, pkg, true)){
    lm_error_set(LM_ERR_PkgNotInstalled);
    return false;
  }

  if(!lm_ctx_database_init(ctx))
    return false;

  lm_pkg_t cur;

  while (lm_database_package_next(ctx->db, &cur)) {
    if(lm_package_depends(&cur, pkg->name)){
      lm_error_set(LM_ERR_PkgBreaks, cur.name);
      lm_database_package_next_free(ctx->db, &cur);
      return false;
    }
  }

  char *path = NULL, *hash = NULL, *fpath = NULL;
  bool in_keep = false, ret = false;
  size_t total = 0, current = 0;

  total = lm_database_files_count(ctx->db, pkg);

  while(lm_database_files_next(ctx->db, pkg, &path, &hash, &in_keep)){
    if(in_keep)
      goto next;

    fpath = join_alloc(ctx->root, path);
    pdebug(__func__, "removing file %s (%s)", fpath, pkg->name);

    if(!exists(fpath)){
      pdebug(__func__, "found file in database, but its not on the file system: %s", fpath);
      goto next;
    }
   
    if(unlink(fpath) < 0){
      pdebug(__func__, "failed to delete file for removing %s: %s", pkg->name, strerror(errno));
      lm_error_set(LM_ERR_PkgFileUnlinkFail, path, strerror(errno));
      goto next;
    }

  next:
    free(fpath);
    fpath = NULL;

    if(NULL != callback && !callback(ctx, pkg, fpath, ++current, total, data))
      goto end;
  }

  if(!lm_database_package_del(ctx->db, pkg)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete %s from the database: %s", pkg->name, lm_strerror());
    lm_error_set(LM_ERR_PkgDatabaseDelFail, suberr);
    free(suberr);
    goto end;
  }

  if(!lm_database_files_del(ctx->db, pkg)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete files of %s from the database: %s", pkg->name, suberr);
    lm_error_set(LM_ERR_PkgFilesDelFail, suberr);
    free(suberr);
    goto end;
  }

  if(!lm_database_changes_del(ctx->db, pkg)){
    char *suberr = lm_strerror_dup();
    pdebug(__func__, "failed to delete changes file for %s: %s", pkg->name, suberr);
    lm_error_set(LM_ERR_PkgChangesDelFail, suberr);
    goto end;
  }

  ret = true;
end: 
  lm_database_package_next_free(ctx->db, &cur);
  lm_database_files_next_free(ctx->db, pkg, &path, &hash, &in_keep);
  return ret;
}

